/*
 * Indie Game Jame 2 - example game "framework"
 * 2004/3/15 checker@d6.com
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
#include <set>
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "IGJ2 - Sleep wit' da Fishes!";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 100.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
// needed for game consistency on different machines
class River_Force : public Global_Force {
public:
	Vec2 m_g;
	River_Force(const Vec2& g) : m_g(g) { }

	void apply_expl(SV& f, const Rigid_Body& rb) const
	{
		if (rb.inv_m != 0.0f) {
			f += SV(m_g * (1.0f/rb.inv_m), 0.0f);
		}
	}
};


#include "../../utils/misc/framerate.h"
Sim simulator;
Drag_Force    global_drag(.05f, .05f);    //a little drag helps with stability
Gravity_Force gravity(Vec2(0, -100.0f)); 
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);
Rigid_Body* floor_body = NULL; 
Rigid_Body* endwall_body = NULL; 
River_Force	  riverFlow(Vec2(5.0f, 0)); 

char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
	float RiverFlow; 
	float UnderwaterDrag; 
	float WaterLevel; 
	float WaterSlope; 
	float WaterDensity; 
	float DensityScale; 
	float DensityRatioClamp; 
	float WaveTimeFreq;
	float WaveSpaceFreq;
	float WaveAmp; 
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;

	float RockDensity; 

	int	  CorpseHitpoints; 
	float CorpseSpawnInterval; 
	float JunkSpawnInterval; 
	float RockSpawnInterval; 


};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;

enum eGameObjectKind
{
	kGOK_None,
	kGOK_Rock,
	kGOK_Junk,
	kGOK_Corpse,
}; 

enum eCollisionBits
{
	kCB_Floor	= 1 << 31, 
	//kCB_CorpseFloor	= 1 << 30, 
	kCB_EndWall = 1 << 29, 
	kCB_Corpse	= 1 << 0,
	kCB_Junk = 1 << 1,
}; 

enum eCollisionMasks
{
	kCM_Rock = kCB_Corpse | kCB_Floor | kCB_Junk, 
	kCM_Corpse = kCB_Corpse | kCB_Floor | kCB_EndWall,
	kCM_Junk = kCB_Junk | kCB_EndWall,
}; 

enum eGameObjectFlags
{
	kGOF_TouchingFloor = 1 << 0, 
}; 



struct game_body
{
	Rigid_Body* pRB;
	real area; 
	real cache_inv_I; 
	GLuint texture; 

	game_body(Rigid_Body* rb)
		: pRB(rb)
	{
		area = 1/(rb->inv_m * rb->cosmetic_density); 
		cache_inv_I = rb->inv_I; 
		texture = 0; 
	}

	~game_body() { }

};

// game stuff
struct game_object {


	void add_body(Rigid_Body* pRB);

	void add_force(Binary_Force* force)
	{
		forceList.push_back(force); 
		simulator.add(*force); 
	}

	void init()
	{
		hitpoints = 1;
		flags = 0; 
	}

    
    game_object( eGameObjectKind k, Rigid_Body *rbInit)
		: kind(k)
    {
		add_body(rbInit); 
    }

	game_object( eGameObjectKind k, Rigid_Body** rbList, int nBodies, Binary_Force** fList = NULL, int nForces = 0)
		: kind(k)
	{
		for (int i = 0; i < nBodies; i++)
		{
			add_body(rbList[i]); 
		}

		for (int i = 0; i < nForces; i++)
		{
			add_force(fList[i]); 
		}
	}

	~game_object()
	{
		for (int i = 0; i < bodyList.size(); i++)
		{
			simulator.remove(*bodyList[i].pRB); 
			delete bodyList[i].pRB; 
		}

		for (int i = 0; i < forceList.size(); i++)
		{
			simulator.remove(*forceList[i]); 
			delete forceList[i]; 
		}
	}

 
	vector<game_body> bodyList; 
	vector<Binary_Force*> forceList; 
 	eGameObjectKind kind; 
	int hitpoints; 
	unsigned long flags; 
	

};

static const char* objectKindNameTable[] = 
{
	"None",
	"Rock",
	"Junk",
	"Corpse",
};

static int objectTextureVariationTable[] = 
{
	0,
	1, // # of rocks
	3, // # of junk objects 
	6  // @NOTE this is the number of corpse heads 
};


GLuint GetTextureByName(const char* name); 

void game_object::add_body(Rigid_Body* pRB)
{
	pRB->m_pAppData = this;  

	game_body body = game_body(pRB);

	int nTextureVariations = objectTextureVariationTable[kind]; 

	if (pRB->name() != "" && nTextureVariations > 0)
	{
		char buf[256]; 
		sprintf(buf,"%s_%s_%02d",objectKindNameTable[kind],pRB->name().c_str(), rand()%nTextureVariations); 

		body.texture = GetTextureByName(buf); 
	}


	if(body.texture == 0 && pRB->name() != "")
	{
		char buf[256]; 
		sprintf(buf,"%s_%s",objectKindNameTable[kind],pRB->name().c_str()); 

		body.texture = GetTextureByName(buf); 
	}

	if (body.texture == 0 && pRB->name() == "" && nTextureVariations > 0)
	{
		char buf[256]; 
		sprintf(buf,"%s_%02d",objectKindNameTable[kind],rand()%nTextureVariations); 

		body.texture = GetTextureByName(buf); 
	}

	if (body.texture == 0)
		body.texture = GetTextureByName(objectKindNameTable[kind]); 


	simulator.add(*pRB); 
	bodyList.push_back(body); 
}


static vector<game_object*> objs;
static vector<Binary_Force *> forces;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;

static game_object* cursor_object; 

struct game_state_type
{
	int score; 
	int escaped; 
	float timestamp; 
	float corpseSpawnTimer; 
	float junkSpawnTimer; 
	float rockSpawnTimer; 

	void reset()
	{
		score = 0;
		escaped = 0; 
		timestamp = 0;
		corpseSpawnTimer = 0; 
		junkSpawnTimer = 0;
		rockSpawnTimer = 0; 
	}

	game_state_type()
	{
		reset(); 
	}
}; 

typedef map<std::string,GLuint> texture_dictionary; 

static game_state_type game_state; 

struct game_assets_type
{
	GLuint background; 
	texture_dictionary textures; 

}; 

static game_assets_type game_assets = { 0 }; 

GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    if (!LoadBitmapByExtension(Filename,&result))
		return 0; 

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

GLuint GetTextureByName(const char* name)
{
	texture_dictionary::const_iterator pPair = game_assets.textures.find(name); 

	if (pPair != game_assets.textures.end())
		return pPair->second; 

	char buf[256]; 
	sprintf(buf, "%s.psd",name); 

	GLuint result = LoadTexture(buf);
	game_assets.textures[name] = result;

	return result; 
}

real ComputeWaterLevelAtX(real x); 

void
glWater(int nSteps = 100)
{
	if (nSteps < 1)
		return; 

	glEnable(GL_BLEND); 
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	glBegin(GL_TRIANGLE_STRIP);
	glColor4f(0,0,1,0.3f); 

	float xl = -80;
	float xh = +80;
	float yl = -80; 

	for (int i = 0; i < nSteps; i++)
	{
		float x = xl + (xh - xl)*i/(nSteps-1); 
		glVertex2f(x, ComputeWaterLevelAtX(x)); 
		glVertex2f(x,yl);
	}
	glEnd(); 
}

void draw_background() {
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
//    glOrtho(-1, 1, -1, 1, 0, 1);

	glBindTexture(GL_TEXTURE_2D, game_assets.background);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glColor3f(1, 1, 1);
    float z = -0.5f;
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);
    glVertex3f(-1, +1, z);
    glTexCoord2f(1, 0);
    glVertex3f(+1, +1, z);
    glTexCoord2f(1, 1);
    glVertex3f(+1, -1, z);
    glTexCoord2f(0, 1);
    glVertex3f(-1, -1, z);
    glEnd();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);

    glDisable(GL_TEXTURE_2D);
}


void
glBody1v(Rigid_Body* pRB, Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				static const Vec2 texTable[] = { Vec2(0,0), Vec2(1,0), Vec2(1,1), Vec2(0,1),  }; 
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glTexCoord2f(texTable[i].x(), texTable[i].y()); 
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		// MAHK SEZ: Behold the hack.
		// When solid, we tesselate the *circumscribed* square, so that the texture 
		// has the same dimensions as the circle. 
		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				static const Vec2 texTable[] = { Vec2(0,0), Vec2(1,0), Vec2(1,1), Vec2(0,1), Vec2(0,0) }; 

				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = (Solid) ? 5 : 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glTexCoord2f(0.5f, 0.5f); 
					glVertex2f(center.x(), center.y());
				}

				float Radius = p->r();

				if (Solid) Radius *= sqrt(2.0f); 

				for(int a = 0;a<total_steps;++a)
				{
					float Angle = pRB->O.z() + 2.0f * 3.14159265358979323846f * (0.625f + ((float)a / (float)(total_steps - 1)));
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;

					glTexCoord2f(texTable[a].x(), texTable[a].y()); 

					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(pRB, &l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
		delete objs[i]; 
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }

    forces.clear();
    objs.clear();
}

game_object* LoadAndAddFile(char* fileName, const Vec2& loc, eGameObjectKind kind = kGOK_None)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(fileName, loc.x(),loc.y(), 1.0f/6, Globals.DensityScale, 
                    &rb_list, &num_rb, &force_list, &num_force, &world_body);

	game_object* pObject = new game_object(kind, rb_list, num_rb, force_list, num_force); 

	objs.push_back(pObject); 

	return pObject; 
}

void
LoadTestFile(void)
{
	static const std::string floorName = "floor"; 
	static const std::string endwallName = "endwall"; 

	game_object* pObject = LoadAndAddFile(TestLevelFileName, Vec2::zero());

	for (int i =0 ;i < pObject->bodyList.size(); i++)
	{
		if (pObject->bodyList[i].pRB->name() == floorName)
			floor_body = pObject->bodyList[i].pRB; 
		if (pObject->bodyList[i].pRB->name() == endwallName)
			endwall_body = pObject->bodyList[i].pRB; 
	}



}





void ReleaseCursorObject()
{
	if (!cursor_object)
		return; 

	vector<game_body>& bodyList = cursor_object->bodyList; 

	for (int i = 0; i < bodyList.size(); i++)
	{
		game_body& body = bodyList[i]; 

		body.pRB->inv_m = 1.0/(body.pRB->cosmetic_density * body.area); 
		body.pRB->inv_I = body.cache_inv_I; 
		body.pRB->m_Pbody->m_collision_plane = ~0;
		body.pRB->update_derived_state(); 
	}
	cursor_object = NULL; 
}

void SetCursorObject(game_object* pObject)
{
	if (cursor_object)
		ReleaseCursorObject(); 

	cursor_object = pObject; 

	vector<game_body>& bodyList = cursor_object->bodyList; 

	for (int i = 0; i < bodyList.size(); i++)
	{
		game_body& body = bodyList[i]; 
		body.pRB->inv_m = 0; 
		body.pRB->inv_I = 0; 
		body.pRB->m_Pbody->m_collision_plane = 0;
	}
}

void CreateCorpse()
{
	game_object* pCorpse = LoadAndAddFile("corpse.physdb", Vec2(-45,18), kGOK_Corpse);
	pCorpse->hitpoints = Globals.CorpseHitpoints;  
}

void CreateRock()
{
	real radius = real(rand()%3 + 3); 

	std::auto_ptr<Body> pBody = make_Body_Sphere( radius ); 
//	std::auto_ptr<Body> pBody = make_Body_Box( Vec2(-radius,-radius), Vec2(+radius,+radius) ); 

	pBody->m_collision_plane = kCM_Rock; 
	Rigid_Body* pRB = new Rigid_Body( pBody );

	pRB->cosmetic_density = Globals.RockDensity *  Globals.DensityScale;
	real area = (3.14159 * radius * radius); 
	pRB->inv_m = 1.0/(pRB->cosmetic_density * area); 
	pRB->k_friction = 3.0f; 

	game_object* pObject = new game_object(kGOK_Rock, pRB); 

	objs.push_back ( pObject ); 

	SetCursorObject(pObject); 
}

void CreateJunk(const Vec2& loc)
{
	real radius = real(rand()%3 + 1); 

	std::auto_ptr<Body> pBody = make_Body_Box( Vec2(-radius,-radius), Vec2(+radius,+radius) ); 
	pBody->m_collision_plane = kCM_Junk; 
	Rigid_Body* pRB = new Rigid_Body( pBody, loc ); 

	pRB->cosmetic_density = 0.75 * Globals.DensityScale;
	real area = (3.14159 * radius * radius); 
	pRB->inv_m = 1.0/(pRB->cosmetic_density * area); 
	pRB->k_friction = 0; 

	game_object* pObject = new game_object(kGOK_Junk,pRB);
	objs.push_back ( pObject ); 

}

real ComputeWaterLevelAtX(real x); 

void CreateStartingJunk()
{
	// Add a bunch of floaty crap
	static const int kNumSpheres = 10; 
	for (int i = 0; i < kNumSpheres; i++)
	{
		real x = - 40.0f + 80.f * i/kNumSpheres;
		CreateJunk(	Vec2( x, ComputeWaterLevelAtX(x))); 
	}
}


/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
	PS2Controller.init();


    GlobalProperties = ReadPropertyFromFile("fishes.igj2props"); 
//		CloneProperty(PROP_LOCAL_STRUCT(global_properties));
//    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
//                            GlobalProperties);

	ConvertPropertyToStruct(PROP_LOCAL_STRUCT(global_properties), GlobalProperties, &Globals); 

    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
//    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    InitializeFrameRate(60.0f);
    
    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

	// Load textures 
	game_assets.background = LoadTexture("river_bottom.psd"); 

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
	simulator.add(riverFlow); 
    simulator.add(world_body);

    LoadTestFile();

	CreateStartingJunk(); 
}


real ComputeWaterLevelAtX(real x)
{
	real waterLevel = Globals.WaterLevel; 

	waterLevel += x * Globals.WaterSlope; 

	// Compute sinusoidal water level 
	real theta1 = Globals.WaveTimeFreq * game_state.timestamp;
	real theta2 = Globals.WaveSpaceFreq * x; 

	waterLevel += Globals.WaveAmp * sin( theta1 - theta2);  // Moving wave  
//	waterLevel += Globals.WaveAmp * sin( theta1) * sin( theta2 );  // Standing wave  

	return waterLevel;
}


void FloatAllObjs( float dt )
{
	game_state.timestamp += dt; 
    for(int i = 0; i < objs.size(); ++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

		bool touchingFloor = (obj.flags & kGOF_TouchingFloor) != 0; 
		obj.flags &= ~kGOF_TouchingFloor; 

		if (obj.kind == kGOK_Corpse && !touchingFloor)
			obj.hitpoints = Globals.CorpseHitpoints; 

		vector<game_body>& bodyList = obj.bodyList; 

		for (int j = 0; j < bodyList.size(); j++)
		{
			game_body& body = bodyList[j]; 

			if (body.pRB->inv_m == 0 || body.pRB->cosmetic_density == 0)
				continue; 

			real waterLevel = ComputeWaterLevelAtX(body.pRB->x.x()); 
			

			// How far underwater is our CoM
			real depth = waterLevel - body.pRB->x.y(); 

			// Compute an approximate radius from our area. 
			real area	= body.area; 
			real radius = sqrt( area/3.14159 ) ; 

			// What fraction of the body is underwater 
			real submergedPart = 0.5 + depth/(2*radius);

			submergedPart = max(0, min(submergedPart, 1)); 
			submergedPart *= submergedPart; 

			real densityRatio	= Globals.WaterDensity * Globals.DensityScale / body.pRB->cosmetic_density; 

			// Clamp the density ratio
			densityRatio = min(densityRatio,Globals.DensityRatioClamp); 

			body.pRB->k_gravity	= 1 -  submergedPart * densityRatio;
			body.pRB->k_damp		= (submergedPart > 0) ? Globals.UnderwaterDrag : 1.0f; 
		}
	}
}

//void CreateJunk(const Vec2& loc);


void UpdateSpawns(float dt)
{
	game_state.corpseSpawnTimer -= dt; 
	if (game_state.corpseSpawnTimer < 0)
	{
		game_state.corpseSpawnTimer += Globals.CorpseSpawnInterval; 
		CreateCorpse(); 
	}

	if (game_state.rockSpawnTimer >= 0)
		game_state.rockSpawnTimer -= dt; 
	if (game_state.rockSpawnTimer < 0 && cursor_object == NULL)
	{
		game_state.rockSpawnTimer += Globals.RockSpawnInterval;
		CreateRock(); 
	}


	game_state.junkSpawnTimer -= dt; 
	if (game_state.junkSpawnTimer < 0)
	{
		game_state.junkSpawnTimer += Globals.JunkSpawnInterval;
		real x = -60 + real(rand()%30);
		CreateJunk( Vec2(x, ComputeWaterLevelAtX(x)) ); 
	}
}

void UpdateCursorObject()
{
	if (cursor_object)
	{
		real x = world_mouse.x(); 
		real y = max(ComputeWaterLevelAtX(x) + 30, world_mouse.y()); 

		cursor_object->bodyList[0].pRB->x = Vec2(x,y); 
		cursor_object->bodyList[0].pRB->update_derived_state(); 
	}
}

void RemoveGameObject(game_object* pObject)
{

	vector<game_object*>::iterator ppObj; 
	for (ppObj = objs.begin(); ppObj != objs.end(); ppObj++)
		if (*ppObj == pObject)
		{
			objs.erase(ppObj); 
			break; 
		}

	delete pObject; 
}


void TouchFloor(game_object* pObject)
{
	pObject->flags |= kGOF_TouchingFloor; 

	switch (pObject->kind)
	{
	case kGOK_Rock:
		RemoveGameObject(pObject); 
		break; 

	case kGOK_Corpse:
		pObject->hitpoints--; 

		if (pObject->hitpoints <= 0)
		{
			RemoveGameObject(pObject); 
//			CreateCorpse(); 
			game_state.score++; 
		}

		break; 
	}
}

void TouchEndWall(game_object* pObject)
{
	switch(pObject->kind)
	{
	case kGOK_Corpse:
		game_state.escaped++; 
		// @TODO: exact penatly
//		CreateCorpse(); 
		break; 
	}


	RemoveGameObject(pObject); 
}



inline Rigid_Body* IsHittingBody(Rigid_Body* pA, Rigid_Body* pB, Rigid_Body* pTestBody)
{
	if (pTestBody == pA)
		return pB; 
	if (pTestBody == pB)
		return pA; 

	return NULL; 
}


void DetectBoundaryContacts()
{
	const Contact_List& contacts = simulator.m_contacts; 
	Contact_List::const_iterator pContact; 

	set<game_object*> touchingFloor; 
	set<game_object*> touchingEndwall; 

	for (pContact = contacts.begin(); pContact != contacts.end(); pContact++)
	{
		Rigid_Body* pRBA = (Rigid_Body*)pContact->m_Prb_a;
		Rigid_Body* pRBB = (Rigid_Body*)pContact->m_Prb_b;
		Rigid_Body* pOtherBody = NULL; 

		pOtherBody = IsHittingBody(pRBA,pRBB, floor_body); 

		if (pOtherBody) 
			touchingFloor.insert((game_object*)pOtherBody->m_pAppData); 

		pOtherBody = IsHittingBody(pRBA,pRBB, endwall_body); 

		if (pOtherBody) 
			touchingEndwall.insert((game_object*)pOtherBody->m_pAppData); 

	}

	set<game_object*>::iterator ppObject; 

	for (ppObject = touchingFloor.begin(); ppObject != touchingFloor.end(); ppObject++)
		TouchFloor(*ppObject); 

	for (ppObject = touchingEndwall.begin(); ppObject != touchingEndwall.end(); ppObject++)
		TouchEndWall(*ppObject); 


}



/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

    gravity.m_g.set_y(Globals.Gravity);
	riverFlow.m_g.set_x(Globals.RiverFlow); 
    
	PS2Controller.sample();
	Vec2 LeftStick = PS2Controller.left_stick();
	Vec2 RightStick = PS2Controller.right_stick();

	/**************************** step the simulator ****************************/   

    if (spring)                             //update
        spring->m_a = Handle(world_body, world_mouse);

	DetectBoundaryContacts(); 
	FloatAllObjs(1/GetFrameRate()); 
	UpdateSpawns(1/GetFrameRate()); 
	UpdateCursorObject(); 

    int const NumberOfSimSteps = 1;
    for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
        // step as many times as you like
        simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
        // do any game stuff you want here (like per-step collision response, etc.)
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());


    /**************************** render everything *****************************/
   
    // lots of gl stuff
//    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

	draw_background(); 

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

	/*
    // draw axes
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();
	*/ 

    // draw mouse
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	
	// iterate all the bodies, rendering them
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

		if (obj.kind == kGOK_None)
			continue; 

		for (int j = 0; j < obj.bodyList.size(); ++j)
		{
			GLuint texture = obj.bodyList[j].texture; 
			glBindTexture(GL_TEXTURE_2D, obj.bodyList[j].texture);

			if (texture > 0)
			{
				glDisable(GL_DEPTH_TEST);
				glEnable(GL_TEXTURE_2D);
				glDisable(GL_CULL_FACE);
				glColor3f(1, 1, 1);
			}
			else
			{
			    glDisable(GL_TEXTURE_2D);
			    glColor3f(0,1,0);
			}

	        glBody1v(obj.bodyList[j].pRB,obj.bodyList[j].pRB->m_Pbody.get(), true);
		}
    }

	glDisable(GL_TEXTURE_2D);


    glColor3f(0,0,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

		if (obj.kind == kGOK_None)
			continue; 

		for (int j = 0; j < obj.bodyList.size(); ++j)
		{
			GLuint texture = obj.bodyList[j].texture; 
			if (texture != 0)
				continue; 
	        glBody1v(obj.bodyList[j].pRB, obj.bodyList[j].pRB->m_Pbody.get(), false);
		}
    }
    
    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        Binary_Force *Force = forces[i];

        glForce1v(Force);
    }
    glForce1v(spring); 

	glWater(); 

    glDisable(GL_BLEND);

	glColor3f(0, 0, 1);
	glBegin(GL_LINES);
	float StickScale = 10.0f;
	glVertex2f(0, 0);
	glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
	glVertex2f(0, 0);
	glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
	glEnd();

    glColor3f(1, 1, 1);
    char buf[100];
	sprintf(buf,"Sunk: %d  Escaped: %d",game_state.score,game_state.escaped); 
    DrawString(100,550,buf);

	sprintf(buf,"wm: %f,%f", world_mouse.x(), world_mouse.y());  
    DrawString(50,50,buf);



	char *ButtonStrings[] =
	{
		"BUTTON_RIGHT_D",
		"BUTTON_RIGHT_U",
		"BUTTON_RIGHT_L",
		"BUTTON_RIGHT_R",
		"BUTTON_SHOULDER_R1",
		"BUTTON_SHOULDER_R2",
		"BUTTON_SHOULDER_L1",
		"BUTTON_SHOULDER_L2",
		"BUTTON_RSTICK",
		"BUTTON_LSTICK",
		"BUTTON_START",
		"BUTTON_SELECT",
	};

	int Y = 70;
	for(int ButtonIndex = 0;
		ButtonIndex < Joystick::NUM_BUTTONS;
		++ButtonIndex)
	{
		if(PS2Controller.button_is_down(ButtonIndex))
		{
			DrawString(50,(float)Y,ButtonStrings[ButtonIndex]);
			Y += 10;
		}
	}

    DrawTimingBars();
    SwapBuffers(WindowDC);
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

			ReleaseCursorObject(); 

			/*
            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }
			*/ 
        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

//            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }


            break;
        }
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}

	          ClearMouseSpring();
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }

                case 'E': {
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
			if(MovingCamera)
			{
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
				CameraX -= d_world.x();
				CameraY -= d_world.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
